from docutils.parsers.rst.directives.body import Sidebar


class Margin(Sidebar):
    """Goes in the margin to the right of the page."""

    optional_arguments = 1
    required_arguments = 0

    def run(self):
        """Run the directive."""
        if not self.arguments:
            self.arguments = [""]
        nodes = super().run()
        nodes[0].attributes["classes"].append("margin")

        # Remove the "title" node if it is empty
        if not self.arguments:
            nodes[0].children.pop(0)
        return nodes



"""Generate compiled static translation assets for Sphinx."""

import json
import os
from pathlib import Path
import subprocess

# In case the smodin.io code is different from the Sphinx code
RENAME_LANGUAGE_CODES = {
    "zh-cn": "zh_CN",
    "zh-tw": "zh_TW",
}


def convert_json(folder=None):
    """Convert JSON translations into .mo/.po files for Sphinx.
    folder:
        the source folder of the JSON translations. This function will put the
        compiled .mo/.po files in a specific folder relative to this source
        folder. This parameter is just provided to make testing easier.
    """
    # Raw translation JSONs that are hand-edited
    folder = folder or Path(__file__).parent / "assets" / "translations"
    # Location of compiled static translation assets
    out_folder = folder / ".." / ".." / "theme" / "sphinx_book_theme" / "static"

    # compile po
    for path in sorted((folder / "jsons").glob("*.json")):
        data = json.loads(path.read_text("utf8"))
        assert data[0]["symbol"] == "en"
        english = data[0]["text"]
        for item in data[1:]:
            language = item["symbol"]
            if language in RENAME_LANGUAGE_CODES:
                language = RENAME_LANGUAGE_CODES[language]
            out_path = (
                out_folder / "locales" / language / "LC_MESSAGES" / "booktheme.po"  # noqa: E501
            )
            if not out_path.parent.exists():
                out_path.parent.mkdir(parents=True)
            if not out_path.exists():
                header = f"""
msgid ""
msgstr ""
"Project-Id-Version: Sphinx-Book-Theme\\n"
"MIME-Version: 1.0\\n"
"Content-Type: text/plain; charset=UTF-8\\n"
"Content-Transfer-Encoding: 8bit\\n"
"Language: {language}\\n"
"Plural-Forms: nplurals=2; plural=(n != 1);\\n"
"""
                out_path.write_text(header)

            with out_path.open("a") as f:
                f.write("\n")
                f.write(f'msgid "{english}"\n')
                text = item["text"].replace('"', '\\"')
                f.write(f'msgstr "{text}"\n')

    # compile mo
    for path in (out_folder / "locales").glob("**/booktheme.po"):
        print(path)
        subprocess.check_call(
            [
                "msgfmt",
                os.path.abspath(path),
                "-o",
                os.path.abspath(path.parent / "booktheme.mo"),
            ]
        )


if __name__ == "__main__":
    print("[SBT]: Compiling translations")
    convert_json()



from docutils.nodes import Element
from collections.abc import Iterator


def findall(node: Element, *args, **kwargs) -> Iterator[Element]:
    # findall replaces traverse in docutils v0.18
    # note a difference is that findall is an iterator
    impl = getattr(node, "findall", node.traverse)
    return iter(impl(*args, **kwargs))



from sphinx.transforms.post_transforms import SphinxPostTransform
from typing import Any
from docutils import nodes as docutil_nodes
from sphinx import addnodes as sphinx_nodes
from pydata_sphinx_theme.utils import get_theme_options_dict
from .nodes import SideNoteNode
from ._compat import findall


class HandleFootnoteTransform(SphinxPostTransform):
    """Transform footnotes into side/marginnotes."""

    default_priority = 1
    formats = ("html",)

    def run(self, **kwargs: Any) -> None:
        theme_options = get_theme_options_dict(self.app)
        if theme_options.get("use_sidenotes", False) is False:
            return None
        # Cycle through footnote references, and move their content next to the
        # reference. This lets us display the reference in the margin,
        # or just below on narrow screens.
        for ref_node in findall(self.document, docutil_nodes.footnote_reference):
            parent = None
            # Each footnote reference should have a single node it points to via `ids`
            for foot_node in findall(self.document, docutil_nodes.footnote):
                # matching the footnote reference with footnote
                if (
                    len(foot_node.attributes["backrefs"])
                    and foot_node.attributes["backrefs"][0]
                    == ref_node.attributes["ids"][0]
                ):
                    parent = foot_node.parent
                    # second children of footnote node is the content text
                    foot_node_content = foot_node.children[1].children

                    sidenote = SideNoteNode()
                    para = docutil_nodes.inline()
                    # first children of footnote node is the label
                    label = foot_node.children[0].astext()

                    if foot_node_content[0].astext().startswith("{-}"):
                        # marginnotes will have content starting with {-}
                        # remove the number so it doesn't show
                        para.attributes["classes"].append("marginnote")
                        foot_node_content[0] = docutil_nodes.Text(
                            foot_node_content[0].replace("{-}", "")
                        )
                        para.children = foot_node_content

                        sidenote.attributes["names"].append(f"marginnote-role-{label}")
                    else:
                        # sidenotes are the default behavior if no {-}
                        # in this case we keep the number
                        superscript = docutil_nodes.superscript("", label)
                        para.attributes["classes"].append("sidenote")
                        parachildren = [superscript] + foot_node_content
                        para.children = parachildren

                        sidenote.attributes["names"].append(f"sidenote-role-{label}")
                        sidenote.append(superscript)

                    # If the reference is nested (e.g. in an admonition), duplicate
                    # the content node And place it just before the parent container,
                    # so it works w/ margin. Only show one or another depending on
                    # screen width.
                    node_parent = ref_node.parent
                    para_dup = para.deepcopy()
                    # looping to check parent node
                    while not isinstance(
                        node_parent, (docutil_nodes.section, sphinx_nodes.document)
                    ):
                        # if parent node is another container
                        if not isinstance(
                            node_parent,
                            (docutil_nodes.paragraph, docutil_nodes.footnote),
                        ):
                            node_parent.replace_self([para, node_parent])
                            para_dup.attributes["classes"].append("d-n")
                            break
                        node_parent = node_parent.parent

                    ref_node.replace_self([sidenote, para_dup])
                    break
            if parent:
                parent.remove(foot_node)



"""Launch buttons for Binder / Thebe / Colab / etc."""

from pathlib import Path
from typing import Any, Optional
from urllib.parse import urlencode, quote

from docutils.nodes import document
from sphinx.application import Sphinx
from sphinx.locale import get_translation
from sphinx.util import logging
from shutil import copy2

from . import get_repo_parts, get_repo_url


SPHINX_LOGGER = logging.getLogger(__name__)

MESSAGE_CATALOG_NAME = "booktheme"
translation = get_translation(MESSAGE_CATALOG_NAME)


def add_launch_buttons(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: dict[str, Any],
    doctree: Optional[document],
):
    """Builds a binder link and inserts it in HTML context for use in templating.

    This is a ``html-page-context`` sphinx event (see :ref:`sphinx:events`).

    :param pagename: The sphinx docname related to the page
    :param context: A dictionary of values that are given to the template engine,
        to render the page and can be modified to include custom values.
    :param doctree: A doctree when the page is created from a reST documents;
        it will be None when the page is created from an HTML template alone.

    """

    path = app.env.doc2path(pagename)
    extension = Path(path).suffix

    # Don't do anything if no launch provider is configured
    config_theme = app.config["html_theme_options"]
    launch_buttons = config_theme.get("launch_buttons", {})
    if (
        not launch_buttons
        or not _is_notebook(app, context)
        or not any(
            launch_buttons.get(key)
            for key in (
                "binderhub_url",
                "jupyterhub_url",
                "thebe",
                "colab_url",
                "jupyterlite_url",
            )
        )
    ):
        return

    # Grab the header buttons from context as it should already exist.
    header_buttons = context["header_buttons"]

    # Check if we have a markdown notebook, and if so then add a link to the context
    if _is_notebook(app, context) and (
        context["sourcename"].endswith(".md")
        or context["sourcename"].endswith(".md.txt")
    ):
        # Figure out the folders we want
        out_dir = Path(app.outdir)
        build_dir = out_dir.parent
        ntbk_dir = build_dir.joinpath("jupyter_execute")
        sources_dir = out_dir.joinpath("_sources")
        # Paths to old and new notebooks
        path_ntbk = ntbk_dir.joinpath(pagename).with_suffix(".ipynb")
        path_new_notebook = sources_dir.joinpath(pagename).with_suffix(".ipynb")
        # Copy the notebook to `_sources` dir so it can be downloaded
        path_new_notebook.parent.mkdir(exist_ok=True, parents=True)
        copy2(path_ntbk, path_new_notebook)
        context["ipynb_source"] = pagename + ".ipynb"

    # Get repository URL information that we'll use to build links
    repo_url, _ = get_repo_url(context)
    provider_url, org, repo, provider = get_repo_parts(context)
    if org is None and repo is None:
        # Skip the rest because the repo_url isn't right
        return

    branch = _get_branch(config_theme)

    # Construct the extra URL parts (app and relative path)
    notebook_interface_prefixes = {"classic": "tree", "jupyterlab": "lab/tree"}
    notebook_interface = launch_buttons.get("notebook_interface", "classic")
    if notebook_interface not in notebook_interface_prefixes:
        raise ValueError(
            "Notebook UI for Binder/JupyterHub links must be one"
            f"of {tuple(notebook_interface_prefixes.keys())},"
            f"not {notebook_interface}"
        )
    ui_pre = notebook_interface_prefixes[notebook_interface]

    # Check if we have a non-ipynb file, but an ipynb of same name exists
    # If so, we'll use the ipynb extension instead of the text extension
    if extension != ".ipynb" and Path(path).with_suffix(".ipynb").exists():
        extension = ".ipynb"

    # Construct a path to the file relative to the repository root
    book_relpath = config_theme.get("path_to_docs", "").strip("/")
    if book_relpath != "":
        book_relpath += "/"
    path_rel_repo = f"{book_relpath}{pagename}{extension}"

    # Container for launch buttons
    launch_buttons_list = []

    # Now build infrastructure-specific links
    jupyterhub_url = launch_buttons.get("jupyterhub_url", "").strip("/")
    binderhub_url = launch_buttons.get("binderhub_url", "").strip("/")
    colab_url = launch_buttons.get("colab_url", "").strip("/")
    deepnote_url = launch_buttons.get("deepnote_url", "").strip("/")
    # jupyterlite_url could be absolute but without a domain, so we only
    # strip trailing slashes, not leading ones
    jupyterlite_url = launch_buttons.get("jupyterlite_url", "").rstrip("/")

    # Loop through each provider and add a button for it if needed
    if binderhub_url:
        # Any non-standard repository URL should be passed-through raw
        if provider_url not in ["https://github.com", "https://gitlab.com"]:
            # Generic git repository using the full repo URL as a fallback
            url = f"{binderhub_url}/v2/git/{quote(repo_url)}/{branch}"
        elif provider.lower() == "github":
            url = f"{binderhub_url}/v2/gh/{org}/{repo}/{branch}"
        elif provider.lower() == "gitlab":
            # Binder uses %2F for gitlab for some reason
            url = f"{binderhub_url}/v2/gl/{org}%2F{repo}/{branch}"

        url = f"{url}?urlpath={ui_pre}/{path_rel_repo}"
        launch_buttons_list.append(
            {
                "type": "link",
                "text": "Binder",
                "tooltip": translation("Launch on") + " Binder",
                "icon": "_static/images/logo_binder.svg",
                "url": url,
            }
        )

    if jupyterhub_url:
        url_params = urlencode(
            dict(
                repo=repo_url, urlpath=f"{ui_pre}/{repo}/{path_rel_repo}", branch=branch
            ),
            safe="/",
        )
        url = f"{jupyterhub_url}/hub/user-redirect/git-pull?{url_params}"
        launch_buttons_list.append(
            {
                "type": "link",
                "text": "JupyterHub",
                "tooltip": translation("Launch on") + " JupyterHub",
                "icon": "_static/images/logo_jupyterhub.svg",
                "url": url,
            }
        )

    if colab_url:
        if provider.lower() != "github":
            SPHINX_LOGGER.warning(f"Provider {provider} not supported on colab.")
        else:
            url = f"{colab_url}/github/{org}/{repo}/blob/{branch}/{path_rel_repo}"
            launch_buttons_list.append(
                {
                    "type": "link",
                    "text": "Colab",
                    "tooltip": translation("Launch on") + " Colab",
                    "icon": "_static/images/logo_colab.png",
                    "url": url,
                }
            )

    if deepnote_url:
        if provider.lower() != "github":
            SPHINX_LOGGER.warning(f"Provider {provider} not supported on Deepnote.")
        else:
            github_path = f"%2F{org}%2F{repo}%2Fblob%2F{branch}%2F{path_rel_repo}"
            url = f"{deepnote_url}/launch?url=https%3A%2F%2Fgithub.com{github_path}"
            launch_buttons_list.append(
                {
                    "type": "link",
                    "text": "Deepnote",
                    "tooltip": translation("Launch on") + " Deepnote",
                    "icon": "_static/images/logo_deepnote.svg",
                    "url": url,
                }
            )

    if jupyterlite_url:
        jl_ext = launch_buttons.get("jupyterlite_ext", extension).strip()
        jl_rel_repo = f"{book_relpath}{pagename}{jl_ext}"
        url = f"{jupyterlite_url}?path={jl_rel_repo}"
        launch_buttons_list.append(
            {
                "type": "link",
                "text": "JupyterLite",
                "tooltip": "Launch via JupyterLite",
                "icon": "_static/images/logo_jupyterlite.svg",
                "url": url,
            }
        )

    # Add thebe flag in context
    if launch_buttons.get("thebe", False):
        launch_buttons_list.append(
            {
                "type": "javascript",
                "text": translation("Live Code"),
                "tooltip": translation("Launch Thebe"),
                "javascript": "initThebeSBT()",
                "icon": "fas fa-play",
                "label": "launch-thebe",
            }
        )
        context["use_thebe"] = True

    # Add the buttons to header_buttons
    header_buttons.append(
        {
            "type": "group",
            "tooltip": translation("Launch interactive content"),
            "icon": "fas fa-rocket",
            "buttons": launch_buttons_list,
            "label": "launch-buttons",
        }
    )


def _split_repo_url(url):
    """Split a repository URL into an org / repo combination."""
    if "github.com/" in url:
        end = url.split("github.com/")[-1]
        org, repo = end.split("/")[:2]
    else:
        SPHINX_LOGGER.warning(
            f"Currently Binder/JupyterHub repositories must be on GitHub, got {url}"
        )
        org = repo = None
    return org, repo


def _is_notebook(app, context):
    pagename = context["pagename"]
    metadata = app.env.metadata[pagename]
    if "kernelspec" in metadata:
        # Most notebooks will have this
        return True
    elif "ipynb" in context.get("page_source_suffix", ""):
        # Just in case, check for the suffix since some people remove the kernelspec
        return True
    else:
        return False


def _get_branch(config_theme):
    branch = config_theme.get("repository_branch")
    if not branch:
        branch = "master"
    return branch



"""Source file buttons that point to the online repository."""

from pydata_sphinx_theme.utils import get_theme_options_dict
from sphinx.locale import get_translation
from sphinx.util import logging

from . import as_bool, get_repo_url

LOGGER = logging.getLogger(__name__)
MESSAGE_CATALOG_NAME = "booktheme"
translation = get_translation(MESSAGE_CATALOG_NAME)


def add_source_buttons(app, pagename, templatename, context, doctree):
    """Add the source repository buttons."""
    opts = get_theme_options_dict(app)
    header_buttons = context["header_buttons"]
    # If we have a suffix, then we have a source file
    suff = context.get("page_source_suffix")

    # Add HTML context variables that the pydata theme uses that we configure elsewhere
    # For some reason the source_suffix sometimes isn't there even when doctree is
    repo_keywords = [
        "use_issues_button",
        "use_source_button",
        "use_edit_page_button",
        "use_repository_button",
    ]
    for key in repo_keywords:
        opts[key] = as_bool(opts.get(key))

    # Create source buttons for any that are enabled
    if any(opts.get(kw) for kw in repo_keywords):
        # Loop through the possible buttons and construct+add their URL
        repo_buttons = []
        if opts.get("use_repository_button"):
            repo_url, provider = get_repo_url(context)

            repo_buttons.append(
                {
                    "type": "link",
                    "url": repo_url,
                    "tooltip": translation("Source repository"),
                    "text": translation("Repository"),
                    "icon": f"fab fa-{provider.lower()}",
                    "label": "source-repository-button",
                }
            )

        if opts.get("use_source_button") and doctree and suff:
            # We'll reuse this to make action-specific URLs
            provider, edit_url = context["get_edit_provider_and_url"]()
            # Convert URL to a blob so it's for viewing
            if provider.lower() == "github":
                # Use plain=1 to ensure the source text is shown, not rendered
                source_url = edit_url.replace("/edit/", "/blob/") + "?plain=1"
            elif provider.lower() == "gitlab":
                source_url = edit_url.replace("/edit/", "/blob/") + "?plain=1"
            elif provider.lower() == "bitbucket":
                source_url = edit_url.replace("?mode=edit", "")

            repo_buttons.append(
                {
                    "type": "link",
                    "url": source_url,
                    "tooltip": translation("Show source"),
                    "text": translation("Show source"),
                    "icon": "fas fa-code",
                    "label": "source-file-button",
                }
            )

        if opts.get("use_edit_page_button") and doctree and suff:
            # We'll reuse this to make action-specific URLs
            provider, edit_url = context["get_edit_provider_and_url"]()
            repo_buttons.append(
                {
                    "type": "link",
                    "url": edit_url,
                    "tooltip": translation("Suggest edit"),
                    "text": translation("Suggest edit"),
                    "icon": "fas fa-pencil-alt",
                    "label": "source-edit-button",
                }
            )

        if opts.get("use_issues_button"):
            repo_url, provider = get_repo_url(context)
            if provider in ("github", "gitlab"):
                if provider == "github":
                    url = f"{repo_url}/issues/new?title=Issue%20on%20page%20%2F{context['pagename']}.html&body=Your%20issue%20content%20here."  # noqa: E501
                elif provider == "gitlab":
                    url = f"{repo_url}/-/issues/new?issue[title]=Issue%20on%20page%20%2F{context['pagename']}.html&issue[description]=Your%20issue%20content%20here."  # noqa: E501
                repo_buttons.append(
                    {
                        "type": "link",
                        "url": url,
                        "text": translation("Open issue"),
                        "tooltip": translation("Open an issue"),
                        "icon": "fas fa-lightbulb",
                        "label": "source-issues-button",
                    }
                )
            else:
                LOGGER.warning(f"Open issue button not yet supported for {provider}")

        # If we have multiple repo buttons enabled, add a group, otherwise just 1 button
        if len(repo_buttons) > 1:
            header_buttons.append(
                {
                    "type": "group",
                    "tooltip": translation("Source repositories"),
                    "icon": f"fab fa-{provider.lower()}",
                    "buttons": repo_buttons,
                    "label": "source-buttons",
                }
            )
        elif len(repo_buttons) == 1:
            # Remove the text since it's just a single button, want just an icon.
            repo_buttons[0]["text"] = ""
            header_buttons.extend(repo_buttons)



"""Generate metadata for header buttons."""

from sphinx.errors import SphinxError
from sphinx.locale import get_translation
from pydata_sphinx_theme.utils import config_provided_by_user, get_theme_options_dict

from sphinx.util import logging


LOGGER = logging.getLogger(__name__)
MESSAGE_CATALOG_NAME = "booktheme"
translation = get_translation(MESSAGE_CATALOG_NAME)


def as_bool(var):
    """Cast string as a boolean with some extra checks.

    If var is a string, it will be matched to 'true'/'false'
    If var is a bool, it will be returned
    If var is None, it will return False.
    """
    if isinstance(var, str):
        return var.lower() == "true"
    elif isinstance(var, bool):
        return var
    else:
        return False


def get_repo_parts(context):
    """Return the parts of the source repository."""
    for provider in ["github", "bitbucket", "gitlab"]:
        if f"{provider.lower()}_url" in context:
            provider_url = context[f"{provider.lower()}_url"]
            source_user = context[f"{provider.lower()}_user"]
            source_repo = context[f"{provider.lower()}_repo"]
            return provider_url, source_user, source_repo, provider


def get_repo_url(context):
    """Return the provider URL based on what is defined in context."""
    provider_url, user, repo, provider = get_repo_parts(context)
    repo_url = f"{provider_url}/{user}/{repo}"
    return repo_url, provider


def prep_header_buttons(app, pagename, templatename, context, doctree):
    """Prep an empty list that we'll populate with header buttons."""
    context["header_buttons"] = []


def add_header_buttons(app, pagename, templatename, context, doctree):
    """Add basic and general header buttons, we'll add source/launch later."""
    opts = get_theme_options_dict(app)
    pathto = context["pathto"]
    header_buttons = context["header_buttons"]

    # If we have a suffix, then we have a source file
    suff = context.get("page_source_suffix")

    # Download buttons for various source content.
    if as_bool(opts.get("use_download_button", True)) and suff:
        download_buttons = []

        # An ipynb file if it was created as part of the build (e.g. by MyST-NB)
        if context.get("ipynb_source"):
            download_buttons.append(
                {
                    "type": "link",
                    "url": f"{pathto('_sources', 1)}/{context.get('ipynb_source')}",
                    "text": ".ipynb",
                    "icon": "fas fa-code",
                    "tooltip": translation("Download notebook file"),
                    "label": "download-notebook-button",
                }
            )

        # Download the source file
        download_buttons.append(
            {
                "type": "link",
                "url": f"{pathto('_sources', 1)}/{context['sourcename']}",
                "text": suff,
                "tooltip": translation("Download source file"),
                "icon": "fas fa-file",
                "label": "download-source-button",
            }
        )
        download_buttons.append(
            {
                "type": "javascript",
                "javascript": "window.print()",
                "text": ".pdf",
                "tooltip": translation("Print to PDF"),
                "icon": "fas fa-file-pdf",
                "label": "download-pdf-button",
            }
        )

        # Add the group
        header_buttons.append(
            {
                "type": "group",
                "tooltip": translation("Download this page"),
                "icon": "fas fa-download",
                "buttons": download_buttons,
                "label": "download-buttons",
            }
        )

    # Full screen button
    if as_bool(opts.get("use_fullscreen_button", True)):
        header_buttons.append(
            {
                "type": "javascript",
                "javascript": "toggleFullScreen()",
                "tooltip": translation("Fullscreen mode"),
                "icon": "fas fa-expand",
                "label": "fullscreen-button",
                "classes": "pst-navbar-icon",
            }
        )


def update_sourcename(app):
    # Download the source file
    # Sphinx defaults to .txt for html_source_suffix even though the pages almost
    # always are stored in their native suffix (.rst, .md, or .ipynb). So unless
    # the user manually specifies an html_source_suffix, default to an empty string.
    # _raw_config is the configuration as provided by the user.
    # If a key isn't in it, then the user didn't provide it
    if not config_provided_by_user(app, "html_sourcelink_suffix"):
        app.config.html_sourcelink_suffix = ""


def update_context_with_repository_info(app):
    """Update pydata `html_context` options for source from `repository_url`.

    We do this because we use repository_url as one config to define the URL,
    while the PST uses a collection of {provider}_{key} pairs in html_context.
    So here we insert those context variables on our own.
    """
    opts = get_theme_options_dict(app)
    context = app.config.html_context

    # This is the way to give repository info. If it doesn't exist, do nothing.
    repo_url = opts.get("repository_url", "")
    if not repo_url:
        return

    # Check for manually given options first
    branch = opts.get("repository_branch", "")
    provider = opts.get("repository_provider", "")
    relpath = opts.get("path_to_docs", "")
    if branch == "":
        branch = "main"

    # We assume the final two parts of the repository URL are the org/repo
    provider_url, org, repo = repo_url.strip("/").rsplit("/", 2)

    # Infer the provider if it wasn't manually given
    default_provider_urls = {
        "bitbucket": "bitbucket.org",
        "github": "github.com",
        "gitlab": "gitlab.com",
    }

    # If no provider is given, try to infer one from the repo url
    if provider == "":
        for iprov in default_provider_urls.keys():
            if iprov in provider_url.lower():
                provider = iprov
                break

    # If provider is still empty, raise an error because we don't recognize it
    if provider == "":
        raise SphinxError(
            f"Provider not recognized in repository url {repo_url}. "
            "If you're using a custom provider URL, specify `repository_provider`"
        )

    # Update the context because this is what the get_edit_url function uses.
    repository_information = {
        f"{provider}_user": org,
        f"{provider}_repo": repo,
        f"{provider}_version": branch,
        f"{provider}_url": provider_url,
        "doc_path": relpath,
    }
    context.update(repository_information)



from docutils import nodes
from sphinx.application import Sphinx
from typing import Any, cast


class SideNoteNode(nodes.Element):
    """Handles rendering of side/marginnote content text for html outputs.
    Inserts required html to handle both desktop and mobile."""

    def __init__(self, rawsource="", *children, **attributes):
        super().__init__("", **attributes)

    @classmethod
    def add_node(cls, app: Sphinx) -> None:
        add_node = cast(Any, app.add_node)  # has the wrong typing for sphinx<4
        add_node(cls, override=True, html=(visit_SideNoteNode, depart_SideNoteNode))


def visit_SideNoteNode(self, node):
    tagid = node.attributes["names"][0]
    if "marginnote" in tagid:
        self.body.append(
            f"<label for='{tagid}' class='margin-toggle marginnote-label'>"
        )
    else:
        self.body.append(f"<label for='{tagid}' class='margin-toggle'>")
        self.body.append(self.starttag(node, "span"))


def depart_SideNoteNode(self, node):
    tagid = node.attributes["names"][0]
    if "sidenote" in tagid:
        self.body.append("</span>\n\n")
    self.body.append("</label>")
    self.body.append(
        f"<input type='checkbox' id='{tagid}' name='{tagid}' class='margin-toggle'>"
    )



"""A lightweight book theme based on the pydata sphinx theme."""

import hashlib
import os
from pathlib import Path
from functools import lru_cache

from docutils import nodes as docutil_nodes
from sphinx.application import Sphinx
from sphinx.locale import get_translation
from sphinx.util import logging
from pydata_sphinx_theme.utils import get_theme_options_dict

from .directives import Margin
from .nodes import SideNoteNode
from .header_buttons import (
    prep_header_buttons,
    add_header_buttons,
    update_sourcename,
    update_context_with_repository_info,
)
from .header_buttons.launch import add_launch_buttons
from .header_buttons.source import add_source_buttons
from ._compat import findall
from ._transforms import HandleFootnoteTransform

__version__ = "1.4.0"
"""sphinx-book-theme version"""

SPHINX_LOGGER = logging.getLogger(__name__)
DEFAULT_LOG_TYPE = "sphinxbooktheme"
MESSAGE_CATALOG_NAME = "booktheme"


def get_html_theme_path():
    """Return list of HTML theme paths."""
    parent = Path(__file__).parent.resolve()
    theme_path = parent / "theme" / "sphinx_book_theme"
    return theme_path


def add_metadata_to_page(app, pagename, templatename, context, doctree):
    """Adds some metadata about the page that we reuse later."""
    # Add the site title to our context so it can be inserted into the navbar
    if not context.get("root_doc"):
        # TODO: Sphinx renamed master to root in 4.x, deprecate when we drop 3.x
        context["root_doc"] = context.get("master_doc")
    context["root_title"] = app.env.titles[context["root_doc"]].astext()

    # Update the page title because HTML makes it into the page title occasionally
    if pagename in app.env.titles:
        title = app.env.titles[pagename]
        context["pagetitle"] = title.astext()

    # Add a shortened page text to the context using the sections text
    if doctree:
        description = ""
        for section in findall(doctree, docutil_nodes.section):
            description += section.astext().replace("\n", " ")
        description = description[:160]
        context["page_description"] = description

    # Add the author if it exists
    if app.config.author != "unknown":
        context["author"] = app.config.author

    # Translations
    translation = get_translation(MESSAGE_CATALOG_NAME)
    context["translate"] = translation

    # If search text hasn't been manually specified, use a shorter one here
    theme_options = get_theme_options_dict(app)
    if "search_bar_text" not in theme_options:
        context["theme_search_bar_text"] = translation("Search") + "..."


@lru_cache(maxsize=None)
def _gen_hash(path: str) -> str:
    return hashlib.sha1(path.read_bytes()).hexdigest()


def hash_assets_for_files(assets: list, theme_static: Path, context, app):
    """Generate a hash for assets, and append to its entry in context.

    assets: a list of assets to hash, each path should be relative to
         the theme's static folder.

    theme_static: a path to the theme's static folder.

    context: the Sphinx context object where asset links are stored. These are:
        `css_files` and `script_files` keys.
    """
    for asset_path in assets:
        # CSS assets are stored in css_files, JS assets in script_files
        asset_type = "css_files" if asset_path.endswith(".css") else "script_files"
        if asset_type in context:
            # Define paths to the original asset file, and its linked file in Sphinx
            asset_source_path = theme_static / asset_path
            if not asset_source_path.exists():
                SPHINX_LOGGER.warning(
                    f"Asset {asset_source_path} does not exist, not linking."
                )
            # Find this asset in context, and update it to include the digest
            for ii, other_asset in enumerate(context[asset_type]):
                # TODO: eventually the contents of context['css_files'] etc should probably
                #       only be _CascadingStyleSheet etc. For now, assume mixed with strings.
                if getattr(other_asset, "filename", str(other_asset)) != asset_path:
                    continue
                # Take priority from existing asset or use default priority (500)
                priority = getattr(other_asset, "priority", 500)
                # Remove existing asset
                del context[asset_type][ii]
                # Add new asset
                app.add_css_file(
                    asset_path,
                    digest=_gen_hash(asset_source_path),
                    priority=priority,
                )


def hash_html_assets(app, pagename, templatename, context, doctree):
    """Add ?digest={hash} to assets in order to bust cache when changes are made.

    The source files are in `static` while the built HTML is in `_static`.
    """
    assets = ["scripts/sphinx-book-theme.js"]
    # Only append the book theme CSS if it's explicitly this theme. Sub-themes
    # will define their own CSS file, so if a sub-theme is used, this code is
    # run but the book theme CSS file won't be linked in Sphinx.
    if app.config.html_theme == "sphinx_book_theme":
        assets.append("styles/sphinx-book-theme.css")
    hash_assets_for_files(assets, get_html_theme_path() / "static", context, app)


def update_mode_thebe_config(app):
    """Update thebe configuration with SBT-specific values"""
    theme_options = get_theme_options_dict(app)
    if theme_options.get("launch_buttons", {}).get("thebe") is True:
        # In case somebody specifies they want thebe in a launch button
        # but has not activated the sphinx_thebe extension.
        if not hasattr(app.env.config, "thebe_config"):
            SPHINX_LOGGER.warning(
                "Thebe is activated but not added to extensions list. "
                "Add `sphinx_thebe` to your site's extensions list."
            )
            return
        # Will be empty if it doesn't exist
        thebe_config = app.env.config.thebe_config
    else:
        return

    if not theme_options.get("launch_buttons", {}).get("thebe"):
        return

    # Update the repository branch and URL
    # Assume that if there's already a thebe_config, then we don't want to over-ride
    if "repository_url" not in thebe_config:
        thebe_config["repository_url"] = theme_options.get("repository_url")
    if "repository_branch" not in thebe_config:
        branch = theme_options.get("repository_branch")
        if not branch:
            # Explicitly check in case branch is ""
            branch = "master"
        thebe_config["repository_branch"] = branch

    app.env.config.thebe_config = thebe_config


def check_deprecation_keys(app):
    """Warns about the deprecated keys."""

    deprecated_config_list = ["single_page"]
    for key in deprecated_config_list:
        if key in get_theme_options_dict(app):
            SPHINX_LOGGER.warning(
                f"'{key}' was deprecated from version 0.3.4 onwards. See the CHANGELOG for more information: https://github.com/executablebooks/sphinx-book-theme/blob/master/CHANGELOG.md"  # noqa: E501
                f"[{DEFAULT_LOG_TYPE}]",
                type=DEFAULT_LOG_TYPE,
            )


def update_general_config(app, config):
    theme_dir = get_html_theme_path()

    config.templates_path.append(os.path.join(theme_dir, "components"))


def update_templates(app, pagename, templatename, context, doctree):
    """Update template names and assets for page build.

    This is a copy of what the pydata theme does here to include a new section
    - https://github.com/pydata/pydata-sphinx-theme/blob/0a4894fab49befc59eb497811949a1d0ede626eb/src/pydata_sphinx_theme/__init__.py#L173 # noqa: E501
    """
    # Allow for more flexibility in template names
    template_sections = ["theme_footer_content_items"]
    for section in template_sections:
        if context.get(section):
            # Break apart `,` separated strings so we can use , in the defaults
            if isinstance(context.get(section), str):
                context[section] = [
                    ii.strip() for ii in context.get(section).split(",")
                ]

            # Add `.html` to templates with no suffix
            for ii, template in enumerate(context.get(section)):
                if not os.path.splitext(template)[1]:
                    context[section][ii] = template + ".html"


def setup(app: Sphinx):
    # Register theme
    theme_dir = get_html_theme_path()
    app.add_html_theme("sphinx_book_theme", theme_dir)
    app.add_js_file("scripts/sphinx-book-theme.js")

    # Translations
    locale_dir = os.path.join(theme_dir, "static", "locales")
    app.add_message_catalog(MESSAGE_CATALOG_NAME, locale_dir)

    # Events
    app.connect("builder-inited", update_mode_thebe_config)
    app.connect("builder-inited", check_deprecation_keys)
    app.connect("builder-inited", update_sourcename)
    app.connect("builder-inited", update_context_with_repository_info)
    app.connect("html-page-context", add_metadata_to_page)
    app.connect("html-page-context", hash_html_assets)
    app.connect("html-page-context", update_templates)

    # This extension has both theme-like and extension-like features.
    # Themes are initialised immediately before use, thus we cannot
    # rely on an event to set the config - the theme config must be
    # set in setup(app):
    update_general_config(app, app.config)
    # Meanwhile, extensions are initialised _first_, and any config
    # values set during setup() will be overwritten. We must therefore
    # register the `config-inited` event to set these config options
    app.connect("config-inited", update_general_config)

    # Nodes
    SideNoteNode.add_node(app)

    # Header buttons
    app.connect("html-page-context", prep_header_buttons)
    # Bump priority so that it runs after the pydata theme sets up the edit URL func.
    app.connect("html-page-context", add_launch_buttons, priority=501)
    app.connect("html-page-context", add_source_buttons, priority=501)
    app.connect("html-page-context", add_header_buttons, priority=501)

    # Directives
    app.add_directive("margin", Margin)

    # Post-transforms
    app.add_post_transform(HandleFootnoteTransform)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }
